"""
MSeed object to hold and process MSeed data
"""
import os
import re
import traceback
from pathlib import Path
from typing import Dict, Tuple, List

from sohstationviewer.controller.util import validate_file, validate_dir
from sohstationviewer.view.util.enums import LogType

from sohstationviewer.model.general_data.general_data import \
    GeneralData, ThreadStopped, ProcessingDataError
from sohstationviewer.model.general_data.general_data_helper import read_text

from sohstationviewer.model.mseed_data.mseed_helper import \
    retrieve_nets_from_data_dict
from sohstationviewer.model.mseed_data.record_reader_helper import \
    MSeedReadError
from sohstationviewer.model.mseed_data.mseed_reader import MSeedReader

class MSeed(GeneralData):
    """
    read and process mseed file into object with properties can be used to
    plot SOH data, mass position data, waveform data and gaps
    """

    def __init__(self, *args, **kwargs):
        # FROM mseed.mseed.MSEED.__init__
        super().__init__(*args, **kwargs)
        self.nets_by_sta: Dict[str, List[str]] = {}
        self.processing_data()

    def finalize_data(self):
        """
        CHANGED FROM mseed.mseed.MSEED.finalize_data

        This function should be called after all folders finish reading to
            + get nets_by_sta from stream_header_by_key_chan
            + other tasks in super().finalize_data()

        """
        self.distribute_log_text_to_station()
        self.retrieve_nets_from_data_dicts()
        super().finalize_data()

    def read_folder(self, folder: str) -> Tuple[Dict]:
        """
        CHANGED FROM mseed.mseed.MSEED.read_folder

        Read data streams for soh, mass position and waveform.

        :param folder: absolute path to data set folder
        :return waveform_data: waveform data by station
        :return soh_data: soh data by station
        :return mass_pos_data: mass position data by station
        :return gaps: gap list by station
        :return nets_by_sta: netcodes list by station
        """
        if not os.path.isdir(folder):
            raise ProcessingDataError(f"Path '{folder}' not exist")
        count = 0

        total = sum([len(files) for _, _, files in os.walk(folder)])
        invalid_blockettes = False
        not_mseed_files = []
        for path, sub_dirs, files in os.walk(folder):
            try:
                validate_dir(path)
            except Exception as e:
                # skip Information folder
                self.track_info(str(e), LogType.WARNING)
                continue
            for file_name in files:
                if self.creator_thread.isInterruptionRequested():
                    raise ThreadStopped()

                path2file = Path(path).joinpath(file_name)

                if not validate_file(path2file, file_name):
                    continue
                count += 1
                if count % 10 == 0:
                    self.track_info(
                        f'Read {count} files/{total}', LogType.INFO)
                log_text = read_text(path2file)
                if log_text is not None:
                    self.log_texts[path2file] = log_text
                    continue
                reader = MSeedReader(
                    path2file,
                    read_start=self.read_start,
                    read_end=self.read_end,
                    is_multiplex=self.is_multiplex,
                    req_soh_chans=self.req_soh_chans,
                    req_wf_chans=self.req_wf_chans,
                    include_mp123zne=self.include_mp123zne,
                    include_mp456uvw=self.include_mp456uvw,
                    soh_data=self.soh_data,
                    mass_pos_data=self.mass_pos_data,
                    waveform_data=self.waveform_data,
                    log_data=self.log_data,
                    gap_minimum=self.gap_minimum)
                try:
                    reader.read()
                    invalid_blockettes = (invalid_blockettes
                                          or reader.invalid_blockettes)
                except MSeedReadError:
                    not_mseed_files.append(file_name)
                except Exception:
                    fmt = traceback.format_exc()
                    self.track_info(f"Skip file {path2file} can't be read "
                                    f"due to error: {str(fmt)}",
                                    LogType.WARNING)
        if not_mseed_files:
            self.track_info(
                f"Not MSeed files: {not_mseed_files}", LogType.WARNING)
        if invalid_blockettes:
            # This check to only print out message once
            print("We currently only handle blockettes 500, 1000,"
                  " and 1001.")
        self.track_info(
            f'Skipped {total - count} invalid files.', LogType.INFO)

    def retrieve_nets_from_data_dicts(self):
        """
        Going through stations of each data_dict to get all network codes found
            in all channel of a station to add to nets_by_station.
        """
        retrieve_nets_from_data_dict(self.soh_data, self.nets_by_sta)
        retrieve_nets_from_data_dict(self.mass_pos_data, self.nets_by_sta)
        retrieve_nets_from_data_dict(self.waveform_data, self.nets_by_sta)

    def select_key(self) -> str:
        """
        CHANGED FROM mseed.mseed.MSEED:
            + get sta_ids from self.keys
            + add condition if not on_unittest to create unittest for mseed

        :return selected_sta_id: the selected station id from available
            key of stream header.
            + If there is only one station id, return it.
            + If there is more than one, show all ids, let user choose one to
                return.
        """
        self.keys = sorted(list(set(
            list(self.soh_data.keys()) +
            list(self.mass_pos_data.keys()) +
            list(self.waveform_data.keys()) +
            [k for k in list(self.log_data.keys()) if k != 'TEXT']
        )))
        sta_ids = self.keys

        if len(sta_ids) == 0:
            return

        selected_sta_id = sta_ids[0]
        if not self.on_unittest and len(sta_ids) > 1:
            msg = ("There are more than one stations in the given data.\n"
                   "Please select one to display")
            self.pause_signal.emit(msg, sta_ids)
            self.pause()
            selected_sta_id = sta_ids[self.pause_response]

        self.track_info(f'Select Station {selected_sta_id}', LogType.INFO)
        return selected_sta_id

    def distribute_log_text_to_station(self):
        """
        Loop through paths to text files to look for station id in the path.
            + If there is station id in the path, add the content to the
                station id with channel 'TXT'.
            + if station id not in the path, add the content to the key 'TEXT'
                which means don't know the station for these texts.
        """
        for path2file in self.log_texts:
            try:
                file_parts = re.split(rf"{os.sep}|\.", path2file.as_posix())
                sta = [s for s in self.keys if s in file_parts][0]
            except IndexError:
                self.log_data['TEXT'].append(self.log_texts[path2file])
                continue
            if 'TXT' not in self.log_data[sta]:
                self.log_data[sta]['TXT'] = []
            self.log_data[sta]['TXT'].append(self.log_texts[path2file])
